/*
 *  Tcpjunk - TCP protocols testing tool
 *  Contact: oferkv@gmail.com
 *
 *  Tcpjunk is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation; either version 2, or (at your option) any
 *  later version.
 *
 *  Tcpjunk is distributed in the hope that it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 */

#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/wait.h>
#include <netdb.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <fcntl.h>
#include <dirent.h>
#include <netinet/tcp.h>
#include "tcpjunk.h"

void tj_show_help()
  {
  printf("Tcpjunk v%s - TCP protocols testing tool\n", VER);
  puts(
	  "\nGeneral options:\n\
    -c <filename>  session setup file\n\
    -e <directory> use files from current directory sequentially\n\
    -f <directory> use random files from current directory\n\
    -d             force display of session data when using instances\n\
                   or in server mode\n\
    -X             show session data as hex dump\n\
    -v             display warnings and other information\n\
    -h             show this help and exit\n\
    -x             Tcpjunk in GUI mode\n\
    -a             Tcpjunk as a daemon\n\n\
Client options:\n\
    -i <number>    number of instances to run in parallel\n\
    -l <number>    number of loops for each instance (default: forever)\n\
    -s <hostname>  hostname/IP of target\n\
    -P <IP>        use specific source IP address\n\
    -R             use random source IP addresses\n\
    -I <interface> source IP alias interface\n\
    -O <offset>    source IP interface alias offset\n\
    -o <port>      bind to specific source port\n\n\
Server options:\n\
    -r             server mode\n\
    -E <filename>  server Certificate file (default: server.crt)\n\
    -K <filename>  server key file (default: server.key)\n\n\
Connection options:\n\
    -p <port>      TCP port to connect/listen to\n\
    -W             wait for data at first\n\
    -C             do not wait for data at last\n\
    -w <mseconds>  delay between connections (default: none)\n\
    -k <times>     repeat on the same connection (default: 1)\n\
    -t <sec>       connection timeout (default: 2 seconds)\n\
    -T <ms>        reply timeout (default: 100 milliseconds)\n\
    -m <bytes>     send/receive buffer size\n\
    -D <ms>        delay before sending each buffer\n\
    -u             optimize TCP tw_recycle, tw_reuse settings and\n\
                   file descriptor limits\n\
    -S             use SSL v3.0\n\n\
Session tags:\n\
   <sect>[data]</sect>                 delimiters for multiple sections\n\
   <rep [min] [max]>[data]</rep>       data repeater\n\
   <sys>[command]<sys>                 insert output of a command\n\
   <fuzz [type|last] [min] [max]>      insert a fuzzed buffer\n\
   <char [hex code] [len]>             insert a character\n\
   <enc [prefix] [type] [postfix] [padding]>[data]</enc>\n\
                                       insert encoded data\n\
   <count [dec|hex|last] [min] [max]>  insert a counter\n\
   <inst>                              insert client instance number\n\n\
See manual page for more information.\n\
Bug reports and suggestions: <oferkv@gmail.com>");

  exit(EXIT_FAILURE);
  }

void sigchld_handler(int s)
  {
  while (wait(NULL) > 0)
	;
  }

int read_reply(int sock, SSL *ssl, struct tj_data *inst_data)
  {
  int sel_ret, read_ret;
  int read_success = 0;
  char read_buf[SOCK_BUF_LEN];

  /* initialize descriptors set */
  fd_set set;
  FD_ZERO(&set);
  FD_SET(sock, &set);

  /* initialize timeout */
  struct timeval timeout;
  timeout.tv_sec = inst_data->conn_timeout;
  timeout.tv_usec = 0;

  while ((sel_ret = select(sock + 1, &set, NULL, NULL, &timeout)) > 0
	  || (inst_data->use_ssl && SSL_pending(ssl)))
	{
	/* clean buffer */
	memset(read_buf, 0, sizeof(read_buf));

	/* read clear text */
	if (inst_data->use_ssl == 0)
	  {
	  read_ret = recv(sock, read_buf, SOCK_BUF_LEN, MSG_DONTWAIT);
	  if (read_ret < 0)
		{
		tj_error(inst_data, "Error reading data from socket: ",
			strerror(errno), c_red);
		break;
		}
	  }
	/* read using SSL */
	else
	  {
	  read_ret = SSL_read(ssl, read_buf, SOCK_BUF_LEN);
	  if (read_ret < 0)
		{
		tj_error(inst_data, "SSL read ",
			ERR_error_string(ERR_get_error(), NULL), c_red);
		break;
		}
	  }

	/* if no data break */
	if (read_ret == 0)
	  break;

	/* print reply */
	if (inst_data->debug > 0 && inst_data->gui == 0)
	  {
	  if (inst_data->hexdump)
		{
		printf("\n%s", hex_dump(read_buf, read_ret));
		}
	  else
		fwrite(read_buf, sizeof(char), read_ret, stdout);
	  }
	else if (inst_data->gui == 1)
	  gui_show_traffic(read_buf, read_ret, 1, inst_data);

	/* verbose messages */
	if (inst_data->verbose)
	  {
	  char read_len_str[64];
	  sprintf(read_len_str, "Received %d bytes", read_ret);
	  tj_error(inst_data, read_len_str, "", c_blue_hl);
	  }

	if (!read_success)
	  read_success = 1;

	/* reset timeouts */
	timeout.tv_sec = 0;
	timeout.tv_usec = inst_data->reply_timeout;

	}

  if (read_success)
	return 0;
  else
	return 1;

  }

void send_data(int sock, SSL *ssl, struct tj_data *inst_data, char *send_buf,
	int send_buf_len)
  {
  int rc;
  int err;

  if (inst_data->use_ssl == 0)
	{

	/* send junk clear text */
	if (inst_data->seg_size == 0)
	  {
	  rc = send(sock, send_buf, send_buf_len, 0);
	  if (rc < 0)
		tj_error(inst_data, "Error sending data: ", strerror(errno), c_red);
	  }
	else
	  {
	  /* segment size */
	  int loc, seg_len = inst_data->seg_size;
	  for (loc = 0; loc < send_buf_len; loc += seg_len)
		{
		if ((send_buf_len - loc) < seg_len)
		  seg_len = send_buf_len - loc;
		rc = send(sock, &send_buf[loc], seg_len, 0);
		if (inst_data->seg_delay > 0)
		  usleep(inst_data->seg_delay);
		if (rc < 0)
		  {
		  tj_error(inst_data, "Error sending data: ", strerror(errno), c_red);
		  break;
		  }
		}
	  }
	}
  else
	{
	/* send junk using SSL */

	if (inst_data->seg_size == 0)
	  {
	  err = SSL_write(ssl, send_buf, send_buf_len);
	  if ((err) == -1)
		tj_error(inst_data, "SSL write ", ERR_error_string(ERR_get_error(),
			NULL), c_red);
	  }
	else
	  {
	  /* buffer size limit */
	  int loc, seg_len = inst_data->seg_size;
	  for (loc = 0; loc < send_buf_len; loc += seg_len)
		{
		if ((send_buf_len - loc) < seg_len)
		  seg_len = send_buf_len - loc;
		err = SSL_write(ssl, &send_buf[loc], seg_len);
		if (inst_data->seg_delay > 0)
		  usleep(inst_data->seg_delay);
		if ((err) == -1)
		  tj_error(inst_data, "SSL write ", ERR_error_string(ERR_get_error(),
			  NULL), c_red);
		}
	  }
	}

  /* print sent buffer if we are debugging */
  if (inst_data->debug > 0 && inst_data->gui == 0)
	if (inst_data->hexdump)
	  printf("\n%s", hex_dump(send_buf, send_buf_len));
	else
	  fwrite(send_buf, sizeof(char), send_buf_len, stdout);
  else if (inst_data->gui == 1)
	{
	gui_show_traffic(send_buf, send_buf_len, 0, inst_data);
	}
  if (inst_data->verbose)
	{
	char read_len_str[64];
	sprintf(read_len_str, "Sent %d bytes", send_buf_len);
	tj_error(inst_data, read_len_str, "", c_red_hl);
	}

  }

int tj_connect(int sock, struct sockaddr_in sa, struct tj_data *inst_data)
  {
  int ret = 0, sock_opts;

  /* set non-blocking on socket */
  if ((sock_opts = fcntl(sock, F_GETFL, 0)) >= 0)
	ret = fcntl(sock, F_SETFL, sock_opts | O_NONBLOCK);
  if (ret < 0)
	return -1;

  /* buffer size limit */
  if (inst_data->seg_size)
	{
	int flag = 1;
	ret = setsockopt(sock, IPPROTO_TCP, TCP_NODELAY, (char *) &flag,
		sizeof(int));
	if (ret < 0)
	  tj_error(inst_data, "Disabling TCP_NODELAY failed: ", strerror(errno),
		  c_red);
	}

  /* connect */
  ret = connect(sock, (struct sockaddr *) &sa, sizeof(sa));
  if ((ret < 0) && (errno != EINPROGRESS))
	return -1;

  /* initialize descriptors set */
  fd_set set;
  FD_ZERO(&set);
  FD_SET(sock, &set);

  /* initialize timout */
  struct timeval timeout;
  timeout.tv_sec = inst_data->conn_timeout;
  timeout.tv_usec = 0;

  ret = select(sock + 1, NULL, &set, NULL, &timeout);
  if (ret > 0)
	{
	int so_ret, get_ret;
	unsigned int get_len = sizeof(get_ret);

	/* handle errors */
	so_ret = getsockopt(sock, SOL_SOCKET, SO_ERROR, &get_ret, &get_len);
	if (so_ret < 0)
	  {
	  tj_error(inst_data, "System request failed: ", strerror(errno), c_red);
	  return -1;
	  }
	if (get_ret > 0)
	  {
	  errno = get_ret;
	  return -1;
	  }

	}
  else if (ret <= 0)
	{
	errno = ETIMEDOUT;
	return -1;
	}

  /* all was good, restore socket options */
  ret = fcntl(sock, F_SETFL, sock_opts);

  if (inst_data->verbose == 1)
	tj_error(inst_data, "Connected", "", c_black);
  return 0;
  }

void tj_disconnect(int sock, struct tj_data *inst_data, SSL * ssl,
	SSL_CTX * ctx)
  {

  int err = 0;
  if (inst_data->use_ssl == 1)
	{
	if (ssl != NULL)
	  err = SSL_clear(ssl);
	if ((err) == -1)
	  tj_error(inst_data, "SSL Clear ",
		  ERR_error_string(ERR_get_error(), NULL), c_red);
	}

  shutdown(sock, SHUT_RDWR);
  err = close(sock);
  if (err < 0)
	tj_error(inst_data, "Closing socket failed: ", strerror(errno), c_red);
  else if (inst_data->verbose == 1)
	tj_error(inst_data, "Connection closed", "", c_black);

  }

long tj_filesize(FILE *stream)
  {
  long curpos, len;
  curpos = ftell(stream);
  fseek(stream, 0L, SEEK_END);
  len = ftell(stream);
  fseek(stream, curpos, SEEK_SET);
  return len;
  }

void load_session_data(struct tj_data *inst_data)
  {
  /* load session file */
  FILE *fp;
  if ((fp = fopen(inst_data->sess_file_name, "r")) == NULL)
	{
	printf("tcpjunk: can not open session file %s, terminating...\n",
		inst_data->sess_file_name);
	exit(EXIT_FAILURE);
	}

  /* set session length */
  inst_data->sess_data_len = tj_filesize(fp);
  if (inst_data->sess_data_len < 1)
	{
	printf("tcpjunk: the file %s is empty, terminating...\n",
		inst_data->sess_file_name);
	exit(EXIT_FAILURE);
	}

  /* setup file buffer */
  if (inst_data->sess_data)
	free(inst_data->sess_data);
  inst_data->sess_data = tj_malloc(inst_data->sess_data,
	  inst_data->sess_data_len + 1);

  /* read file data */
  fread(inst_data->sess_data, inst_data->sess_data_len, 1, fp);
  fclose(fp);
  }

int check_address(struct tj_data *inst_data, char *addr)
  {

  if (*addr == '\0')
	goto failure;

  /* check address */
  if (inst_data->verbose)
	tj_error(inst_data, "Resolving address", "", c_black);
  struct hostent *hnames;
  hnames = gethostbyname(addr);
  if (!hnames)
	goto failure;
  else
	{
	inst_data->dest_addr = hnames->h_addr_list[0];
	inst_data->dest_addr_len = hnames->h_length;
	if (inst_data->verbose)
	  tj_error(inst_data, "Connecting to ", inet_ntoa(
		  *(struct in_addr *) hnames->h_addr_list[0]), c_black);
	return 1;
	}

  failure:

  tj_error(inst_data, "Invalid address ", addr, c_red);
  if (inst_data->gui == 0)
	exit(EXIT_FAILURE);
  else
	return 0;

  }

int check_src_address(struct tj_data *inst_data, char *addr)
  {
  if (*addr == '\0')
	goto failure;

  /* check address */
  if (inst_data->verbose)
	tj_error(inst_data, "Resolving source address", "", c_black);
  struct hostent *hnames;
  hnames = gethostbyname(addr);
  if (!hnames)
	goto failure;
  else
	{
	inst_data->src_addr = hnames->h_addr_list[0];
	inst_data->src_addr_len = hnames->h_length;
	if (inst_data->verbose)
	  tj_error(inst_data, "Set source address to ", inet_ntoa(
		  *(struct in_addr *) hnames->h_addr_list[0]), c_black);
	return 1;
	}

  failure:

  tj_error(inst_data, "Invalid source address ", addr, c_red);
  if (inst_data->gui == 0)
	exit(EXIT_FAILURE);
  else
	return 0;

  }

int check_port(int port, struct tj_data *inst_data)
  {
  /* check port */
  if (port < 0 || port > 65535)
	{
	tj_error(inst_data, "Invalid port", "", c_red);
	if (inst_data->gui == 0)
	  exit(EXIT_FAILURE);
	else
	  return 0;
	}
  return 1;
  }

void tj_error(struct tj_data *inst_data, char *message, char *param, int color)
  {
  if (inst_data->debug || inst_data->verbose)
	{
	char temp_err[1024];

	/* print to stderr or in GUI */
	if (inst_data->gui == 0)
	  {
	  sprintf(temp_err, "%s %s%s\n", "tcpjunk:", message, param);
	  fprintf(stdout, "%s", temp_err);
	  }
	else
	  {
	  sprintf(temp_err, "%s%s\n", message, param);
	  gui_add_log(temp_err, color);
	  }
	}
  }

void find_sections(struct tj_data *inst_data)
  {
  /* find all send buffers in session data */
  char *sects_buf = inst_data->sess_data;
  char *sects_buf_loc = sects_buf;
  inst_data->sect_idx = 0;
  char *start_sect_tag = "<sect>";
  char *end_sect_tag = "</sect>";

  /* if there are sections, enumerate them */
  while ((sects_buf_loc = memmem(sects_buf_loc, (inst_data->sess_data
	  + inst_data->sess_data_len - sects_buf_loc), start_sect_tag, strlen(
	  start_sect_tag))))
	{

	/* start tag found */
	sects_buf_loc += strlen(start_sect_tag);
	inst_data->sects[inst_data->sect_idx] = sects_buf_loc;

	/* get section end */
	if (!(sects_buf_loc = memmem(sects_buf_loc, sects_buf
		+ inst_data->sess_data_len - sects_buf_loc, end_sect_tag, strlen(
		end_sect_tag))))
	  {
	  /* end tag not found, using all buffer */
	  goto single_sect;
	  }

	/* end tag found */
	inst_data->sects_len[inst_data->sect_idx] = sects_buf_loc
		- inst_data->sects[inst_data->sect_idx];

	inst_data->sect_idx++;
	sects_buf_loc++;
	sects_buf = sects_buf_loc;

	}

  /* just a single section */
  single_sect:

  if (inst_data->sect_idx == 0)
	{
	inst_data->sects[inst_data->sect_idx] = sects_buf;
	inst_data->sects_len[inst_data->sect_idx] = inst_data->sess_data_len;
	inst_data->sect_idx++;
	}
  return;
  }

void set_session_file(struct tj_data *inst_data)
  {

  if (inst_data->auto_file == f_rand)
	{
	/* choose a random file */
	strcpy(inst_data->sess_file_name, inst_data->files_list[rand()
		% inst_data->auto_files_num]);
	}
  else
	{
	/* choose sequential file */
	static int curr_file = 0;
	strcpy(inst_data->sess_file_name, inst_data->files_list[curr_file]);

	curr_file++;
	if (curr_file == inst_data->auto_files_num)
	  curr_file = 0;
	}

  if (inst_data->verbose)
	tj_error(inst_data, "Using session file: ", inst_data->sess_file_name,
		c_black);

  }

void load_file_list(struct tj_data *inst_data, char *auto_file_dir)
  {
  int file_index = 0;
  struct dirent* dfile;
  char file_name[FILENAME_MAX];

  /* open directory */
  DIR *cur_dir = opendir(auto_file_dir);
  if (!cur_dir)
	{
	char temp_err[1024];
	sprintf(temp_err, "tcpjunk: error opening directory %s - ", auto_file_dir);
	tj_error(inst_data, temp_err, strerror(errno), c_red);
	exit(EXIT_FAILURE);
	}

  /* get number of files */
  while ((dfile = readdir(cur_dir)) != NULL)
	{
	if (dfile->d_type != DT_DIR)
	  file_index++;
	}
  inst_data->files_list = (char **) malloc(file_index
	  * sizeof(*inst_data->files_list));
  if (inst_data->files_list == NULL)
	{
	printf("tcpjunk: memory allocation failed, terminating...\n");
	exit(EXIT_FAILURE);
	}

  /* store file names */
  int i = 0;
  rewinddir(cur_dir);
  while (((dfile = readdir(cur_dir)) != NULL))
	{
	if (dfile->d_type != DT_DIR)
	  {
	  sprintf(file_name, "%s%s", auto_file_dir, dfile->d_name);
	  inst_data->files_list[i] = strdup(file_name);

	  if (inst_data->files_list[i] == NULL)
		{
		printf("tcpjunk: memory allocation failed, terminating...\n");
		exit(EXIT_FAILURE);
		}
	  i++;
	  }
	}

  /* close */
  if (closedir(cur_dir) == -1)
	{
	perror("tcpjunk: error closing directory ");
	exit(EXIT_FAILURE);
	}
  inst_data->auto_files_num = file_index;

  }

void set_src_ip(struct tj_data *inst_data, char *bind_if_name, char *new_src_ip)
  {
  /* setup IP on alias */
  in_addr_t ip = inet_addr(new_src_ip);
  int sock = socket(AF_INET, SOCK_DGRAM, 0);
  struct ifreq ifr;

  strcpy(ifr.ifr_name, bind_if_name);
  ifr.ifr_addr.sa_family = AF_INET;
  ((struct sockaddr_in *) &ifr.ifr_addr)->sin_addr.s_addr = ip;
  if (ioctl(sock, SIOCSIFADDR, &ifr) < 0)
	tj_error(inst_data, "Error setting IP address: ", strerror(errno), c_red);

  close(sock);

  }

char *tj_malloc(char *buf, int buf_len)
  {
  buf = malloc(buf_len);
  if (buf == NULL)
	{
	printf("tcpjunk: memory allocation failed, terminating...\n");
	exit(EXIT_FAILURE);
	}
  else
	return buf;
  }

char *tj_realloc(char *buf, int buf_len)
  {
  buf = realloc(buf, buf_len);
  if (buf == NULL)
	{
	printf("tcpjunk: memory allocation failed, terminating...\n");
	exit(EXIT_FAILURE);
	}
  else
	return buf;
  }

void cleanup(int sig)
  {
  printf("\ntcpjunk: terminating...\n");
  exit(EXIT_SUCCESS);
  }

char *hex_dump(char *buf, int buf_len)
  {
  int buf_loc;
  int asc_loc, asc_offset;
  int line_loc = 0;
  static char *xbuf;
  if (xbuf)
	free(xbuf);
  xbuf = tj_malloc(xbuf, buf_len * 10 + 80);
  memset(xbuf, 0, buf_len * 10);
  int xbuf_loc = 0;

  for (buf_loc = 0; buf_loc < buf_len; buf_loc++)
	{
	line_loc++;

	/* start of line, print offset */
	if (line_loc == 1)
	  xbuf_loc += sprintf(xbuf + xbuf_loc, "%08x   ", buf_loc);

	/* print hex char */
	xbuf_loc += sprintf(xbuf + xbuf_loc, "%02x ", (unsigned char) buf[buf_loc]);

	/* print a space if in the middle */
	if (line_loc == 8)
	  xbuf_loc += sprintf(xbuf + xbuf_loc, " ");

	/* end of line or data */
	if (line_loc == 16 || buf_loc == buf_len - 1)
	  {
	  /* do some padding */
	  if (buf_loc == buf_len - 1 && line_loc != 16)
		{
		int temp_loc = line_loc;
		if (temp_loc < 8)
		  xbuf_loc += sprintf(xbuf + xbuf_loc, " ");

		while (temp_loc <= 15)
		  {
		  xbuf_loc += sprintf(xbuf + xbuf_loc, "   ");
		  temp_loc++;
		  }
		}

	  /* print a space between hex data and ASCII */
	  xbuf_loc += sprintf(xbuf + xbuf_loc, "  ");

	  /* print ASCII */
	  asc_offset = 0;
	  for (asc_loc = buf_loc - line_loc + 1; asc_loc <= buf_loc; asc_loc++)
		{
		if (isprint(buf[asc_loc]))
		  xbuf_loc += sprintf(xbuf + xbuf_loc, "%c", buf[asc_loc]);
		else
		  xbuf_loc += sprintf(xbuf + xbuf_loc, ".");
		asc_offset++;
		if (asc_offset == 8)
		  xbuf_loc += sprintf(xbuf + xbuf_loc, " ");
		}

	  /* new line */
	  xbuf_loc += sprintf(xbuf + xbuf_loc, "\n");
	  line_loc = 0;
	  }
	}

  return xbuf;
  }

